#include <dis.hxx>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#ifndef WIN32
#include <sys/wait.h>
#include <unistd.h>
#else
#include <process.h>
#endif
#include <ctype.h>
#include "mcserver.h"

char Center[32];
DLList jobList;

#ifndef WIN32
char outfile[] = "/tmp/mcserver.dat";
#else
char outfile[] = "c:\mcserver.dat";
//char outfile[] = "%TMP%\mcserver.dat";
#endif

extern char **environ;
int my_system(char *);

class Job: public DLLItem, public DimTimer
{
  DimService *jobParams;
  JobService jobService;
  char job_state[132];
  int job_alarm;

  void timerHandler()
  {
    delete(this);
  }

public:
  Job(JobService *ptr, int size)
  {
    char name[132], format[132];
    memcpy(&jobService, ptr, size);
    sprintf(name,"MCCONTROL/%s/PARAMS",jobService.job_id);
    paramFormat(format);

    jobParams = new DimService(name,format,&jobService.job_params,
			       sizeof(mccontrol_params));
  }
  ~Job()
  {
    delete jobParams;
  }
  char *getId() {return jobService.job_id;} 
  void update(JobService *ptr, int size)
  {
    memcpy(&jobService, ptr, size);
    jobParams->updateService();
  }
  void finish()
  {
    strcpy(jobService.job_params.state,"Killed");
    jobParams->updateService();
  }
    
};

class JobDeleteCmd: public DimCommand
{
  char batchId[128];

  void commandHandler();
public:
  JobDeleteCmd(char *name, char *format): DimCommand(name, format) {};
};

void JobDeleteCmd::commandHandler()
{
  Job *job;
  char *job_id;

  job_id = getString();
  job = (Job *)jobList.getHead();
  while(job)
    {
      if(!strcmp(job->getId(),job_id))
	{
	  job->finish();
	  jobList.remove(job);
	  job->start(2);
	  break;
	}
      job = (Job *)jobList.getNext();
    }
}

class JobCmd: public DimCommand
{
  void commandHandler();
public:
  JobCmd(char *name, char *format): DimCommand(name, format) {};
};

void JobCmd::commandHandler()
{
  JobService *ptr;
  int size;
  Job *job;
  int found = 0;

  size = getSize();
  ptr = (JobService *)getData();
  job = (Job *)jobList.getHead();
  while(job)
    {
      if(!strcmp(job->getId(),ptr->job_id))
	{
	  found = 1;
	  job->update(ptr, size);
	  if(!strcmp(ptr->job_params.state, "Finished"))
	    {
	      jobList.remove(job);
	      job->start(2);
	    }
	  break;
	}
      job = (Job *)jobList.getNext();
    }
  if(!found)
    {
      job = new Job(ptr, size);
      jobList.add(job);
    }
}

class FileCmd: public DimCommand
{
  char *infile;
  DimService *jobFile;

  void commandHandler();
public:
  FileCmd(char *name, char *format): DimCommand(name, format) 
  {
    char auxname[132];

    infile = new char[10];
    strcpy(infile,"empty");
    sprintf(auxname,"%s/MCCONTROL/FILE",Center);
    jobFile = new DimService(auxname, "C", infile, strlen(infile)+1);
  };
};

void FileCmd::commandHandler()
{
  char *str;
  char filename[256];
  struct stat buf;
  FILE *fd;
  int ret, size, index, sz;

  str = getString();
  unlink(outfile);
  if(!strncmp(str,"CMD ",4))
    {
      strcpy(filename,&str[4]);
#ifndef WIN32
 /*     strcat(filename," >& "); */
          strcat(filename," > ");
	  strcat(filename,outfile);
	  strcat(filename,"< /dev/null");
#else
      strcat(filename," > ");
	  strcat(filename,outfile);
  /*    strcat(filename," 2>&1 "); */
#endif
      my_system(filename);
      strcpy(filename,outfile);
    }
  else
    strcpy(filename,str);

  delete infile;
  ret = stat(filename, &buf);
  if(ret == -1)
    {
      infile = new char[20];
      strcpy(infile,"File does not exist");
    }
  else
    {
      size = buf.st_size;
      infile = new char[size +1];
      fd = fopen(filename, "r");
      index = 0;
      while(!feof(fd) && size)
	{
	  sz = fread(&infile[index], 1, 512, fd);
	  size -= sz;
	  index += sz;
	  if(!sz)
	    break;
	}
      fclose(fd);
      infile[index] = '\0';
    }
  jobFile->updateService(infile, strlen(infile)+1);
}

#ifndef WIN32

int my_system (char *command)
{
    int pid, status, ret, n = 0;

    if (command == 0)
        return 1;
    pid = fork();
    if (pid == -1)
        return -1;
    if (pid == 0) {
        char *argv[4];
        argv[0] = "sh";
        argv[1] = "-c";
        argv[2] = command;
        argv[3] = 0;
        execve("/bin/sh", argv, environ);
        exit(127);
    }
    do {
      ret = waitpid(pid, &status,WNOHANG);
      if(ret == -1)
	break;
      usleep(100000);
      n++;
    } while(n < 1200);
/*
    if(ret != -1)
      {
	cout << "killing : "<< pid << endl;
	kill(pid, SIGINT);
      }
*/
    return status;
}

#else
int my_system (char *command)
{
	/* return system(command); */
}
 /*
int my_system (char *command)
{
    int pid, status, ret, n = 0;
    char *argv[4];
	HANDLE handle;

    if (command == 0)
        return 1;
    argv[0] = "cmd";
    argv[1] = "/c";
    argv[2] = command;
    argv[3] = 0;
  	pid = spawnvp(_P_NOWAIT, "cmd", argv); 
	do {
		handle = OpenProcess(SYNCHRONIZE, 0, pid);
		ret = WaitForSingleObject(handle,100);
		CloseHandle(handle);
		if(ret != WAIT_TIMEOUT)
			break;
	    n++;
    } while(n < 1200);
    return status;
}
*/

#endif

int getCenterId()
{
  int i;

  for(i = 0; i < strlen(Center); i++)
    Center[i] = toupper(Center[i]);
  for(i = 0; i < MAX_MCCENTERS; i++)
    {
      if(MCCenterNames[i][0])
	{
	  if(!strcmp(Center, MCCenterNames[i]))
	    {
	      // KGG; work around for id>=20 translate ->101
	      if(i>=20)  return i+81;
	      else       return i;
	    }
	}
    }
  return -1;
}


int main(int argc, char **argv)
{
  char name[132], format[132];
  JobCmd *jobCmd;
  JobDeleteCmd *jobDeleteCmd;
  FileCmd *fileCmd;
  int i;

  //  DimServer::setDnsNode("alpc18.cern.ch");
  if(argc != 2)
  {
    cout << "Usage: mcserver <center>" << endl;
    cout << "    Where <center> can be:" << endl;
    for(i = 0; i < MAX_MCCENTERS; i++)
      {
	if(MCCenterNames[i][0])
	  {
	    cout << "      " << MCCenterNames[i] << endl;
	  }
      }
    exit(0);
  }
  strcpy(Center, argv[1]);
  if(getCenterId() == -1)
    {
      cout << "Unknown Center: " << Center << ", Exiting..." << endl;
      exit(0);
    }

  sprintf(name,"%s/MCCONTROL/JOB",Center);
  jobFormat(format);
  jobCmd = new JobCmd(name, format);

  sprintf(name,"%s/MCCONTROL/DELETEJOB",Center);
  jobDeleteCmd = new JobDeleteCmd(name, "C");
  
  sprintf(name,"%s/MCCONTROL/GETFILE",Center);
  fileCmd = new FileCmd(name, "C");
  
  sprintf(name,"%s/MCCONTROL",Center);
  DimServer::start(name);
  while(1)
    {
      pause();
    }
}
